# MIT License

# Copyright (c) 2017 liukuang

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# https://github.com/kuangliu/pytorch-cifar

'''ResNet in PyTorch.

For Pre-activation ResNet, see 'preact_resnet.py'.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

from path_learning.utils.log import get_logger

logger = get_logger("resnet_leaky")


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, negative_slope=0.01, use_group_norm=False):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        if use_group_norm:
            self.bn1 = nn.GroupNorm(32, planes)
        else:
            self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        if use_group_norm:
            self.bn2 = nn.GroupNorm(32, planes)
        else:
            self.bn2 = nn.BatchNorm2d(planes)
        self.negative_slope = negative_slope

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.leaky_relu(self.bn1(self.conv1(x)), negative_slope=self.negative_slope)
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.leaky_relu(out, negative_slope=self.negative_slope)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, input_size=32, num_classes=10, negative_slope=0.01, use_group_norm=False,
                 **kwargs):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        if use_group_norm:
            self.bn1 = nn.GroupNorm(32, 64)
        else:
            self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1, negative_slope=negative_slope)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2, negative_slope=negative_slope)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2, negative_slope=negative_slope)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2, negative_slope=negative_slope)
        self.negative_slope = negative_slope

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.linear = nn.Linear(512*block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride, negative_slope, use_group_norm=False):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride, negative_slope=negative_slope,
                                use_group_norm=use_group_norm))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.leaky_relu(self.bn1(self.conv1(x)), negative_slope=self.negative_slope)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avgpool(out)

        out = out.view(out.size(0), -1)
        out = self.linear(out)

        return out


class ResNetFeatures(ResNet):
    def __init__(self, block, num_blocks, input_size=32, num_classes=10, negative_slope=0.01, use_group_norm=False):
        super(ResNetFeatures, self).__init__(block, num_blocks, input_size=input_size, num_classes=num_classes,
                                             negative_slope=negative_slope, use_group_norm=use_group_norm)
        self.in_planes = 64
        self.block = block
        self.num_classes = num_classes
        self.out_features = self.forward(torch.zeros(1, 3, 32, 32)).size(1)
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        if use_group_norm:
            self.logger.info("USING GROUPNORM")
            self.bn1 = nn.GroupNorm(32, 64)
        else:
            self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1, negative_slope=negative_slope,
                                       use_group_norm=use_group_norm)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2, negative_slope=negative_slope,
                                       use_group_norm=use_group_norm)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2, negative_slope=negative_slope,
                                       use_group_norm=use_group_norm)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2, negative_slope=negative_slope,
                                       use_group_norm=use_group_norm)
        self.negative_slope = negative_slope

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, x):
        out = F.leaky_relu(self.bn1(self.conv1(x)), negative_slope=self.negative_slope)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avgpool(out)

        out = out.view(out.size(0), -1)
        return out


def ResNet18(input_size=32, num_classes=10, negative_slope=0.01, use_groupnorm=False):
    return ResNet(BasicBlock, [2, 2, 2, 2], input_size=input_size, num_classes=num_classes,
                  negative_slope=negative_slope, use_group_norm=use_groupnorm)


def ResNet18Features(input_size=32, num_classes=10, negative_slope=0.01, use_groupnorm=False):
    return ResNetFeatures(BasicBlock, [2, 2, 2, 2], input_size=input_size, num_classes=num_classes,
                          negative_slope=negative_slope, use_group_norm=use_groupnorm)
